@@ -8,6 +8,8 @@ import glob
 import os
 from pathlib import Path
 import platform
+import shutil
+import signal
 import subprocess
 import sys
 import tarfile
@@ -16,7 +18,7 @@ import textwrap
 import time
 from urllib.request import urlopen
 
-__version__ = "0.98"
+__version__ = "0.99"
 
 DOCKER_NS = "satex"
 REGISTRY_URL = "https://github.com/sat-heritage/docker-images/releases/download/list/list.tgz"
@@ -31,6 +33,9 @@ def error(msg):
     print(msg, file=sys.stderr)
     sys.exit(1)
 
+def warn(msg):
+    print("\033[1;33m! %s\033[0m" % msg, file=sys.stderr)
+
 def info(msg):
     print("\033[92m+ %s\033[0m" % msg, file=sys.stderr)
 
@@ -127,7 +132,10 @@ class ImageManager(object):
         self.setup = repo.setup[self.entry].copy()
         self.setup.update(self.setup.get(self.solver, {}))
         self.registry = repo.registry[self.entry][self.solver]
-
+        self.vars = {
+            "SOLVER": self.solver,
+            "SOLVER_NAME": self.solver_name,
+        }
     @property
     def solver_name(self):
         return self.registry["name"]
@@ -137,18 +145,24 @@ class ImageManager(object):
 
 
 def get_list(args):
-    images = Repository(args).names
-    assert len(images), "No matching images!"
+    images = Repository(args).images
+    if(len(images) == 0):
+        sys.exit("No matching images!")
     return images
 
 def print_list(args):
     for image in get_list(args):
         print(image)
 
-
 _info_first = ["name", "version", "authors", "base_from"]
-_info_last = ["comment", "comments"]
+_info_last = ["download_url", "status", "comment", "comments"]
 _info_ignore = {"call"}
+_info_label = {
+    "base_from": "Environment",
+    "download_url": "Download URL",
+    "gz": "Gzip input",
+}
+_info_nowrap = {"download_url"}
 def print_info(args):
     repo = Repository(args)
     for name in repo.images:
@@ -162,21 +176,21 @@ def print_info(args):
         key_width = 0
         info = []
         for key in keys:
+            if key in image.setup:
+                value = image.setup[key].format(**image.vars)
+            elif key in image.registry:
+                value = image.registry[key]
+            else:
+                continue
+            name = _info_label.get(key, key.title())
             if key == "base_from":
-                name = "Environment"
-                value = image.setup[key]
                 builder_base = image.setup.get("builder_base")
                 if builder_base:
                     if builder_base == value:
                         value += " (same as build)"
                     else:
                         value += f" (build: {builder_base})"
-            else:
-                if key not in image.registry:
-                    continue
-                name = key.title()
-                value = image.registry[key]
-            if key in ["args", "argsproof"]:
+            elif key in ["args", "argsproof"]:
                 name = "Call"
                 if key == "argsproof":
                     name += " (proof)"
@@ -184,7 +198,7 @@ def print_info(args):
             else:
                 value = str(value)
             key_width = max(len(name), key_width)
-            info.append((name,value))
+            info.append({"key": key, "name": name, "value": value})
 
         key_width += 2
         line_width = key_width + 70
@@ -196,13 +210,17 @@ def print_info(args):
             color = 33
         print(f"{DOCKER_NS}/\033[1;{color}m{image.name}\033[0m")
         print("-"*line_width)
-        for (key, value) in info:
-            key = f"{key}: "
-            for p in value.splitlines():
-                for line in textwrap.wrap(p):
-                    print("{0:{key_width}}{1}".format(key, line,
-                        key_width=key_width))
-                    key = ""
+        for d in info:
+            name = f"{d['name']}: "
+            if d["key"] in _info_nowrap:
+                print("{0:{key_width}}{1}".format(name, d["value"],
+                            key_width=key_width))
+            else:
+                for p in d["value"].splitlines():
+                    for line in textwrap.wrap(p):
+                        print("{0:{key_width}}{1}".format(name, line,
+                            key_width=key_width))
+                        name = ""
         print()
 
 #
@@ -265,7 +283,10 @@ def prepare_image(args, docker_argv, image):
 _docker_opts = []
 def docker_runs(args, images, docker_args=(), image_args=()):
     docker_argv = check_docker()
-    argv = ["run", "--rm"]
+    container_id = f"satex{os.getpid()}"
+    argv = ["run", "--name", container_id, "--rm"]
+    if hasattr(args, "timeout"):
+        argv += ["-e", f"TIMEOUT={args.timeout}"]
     for opt in _docker_opts:
         if getattr(args, opt) is not None:
             val = getattr(args, opt)
@@ -285,9 +306,21 @@ def docker_runs(args, images, docker_args=(), image_args=()):
         if args.pretend:
             print(" ".join(cmd))
         else:
+            def killer(s,f):
+                warn("Killing solver...")
+                argv = docker_argv + ["kill", container_id]
+                subprocess.run(argv, stdout=subprocess.DEVNULL)
+            signal.signal(signal.SIGINT, killer)
             info(" ".join(cmd))
-            ret = subprocess.call(cmd)
-            assert ret == 0 or 10 <= ret <= 20, f"Solver failed! ({ret})"
+            ret = subprocess.run(cmd).returncode
+            signal.signal(signal.SIGINT, signal.SIG_DFL)
+            if ret == 124:
+                if args.fail_if_timeout:
+                    raise subprocess.TimeoutExpired(image, args.timeout)
+                else:
+                    warn(f"{image} timeout")
+            else:
+                assert ret == 0 or 10 <= ret <= 20, f"Solver failed! ({ret})"
     if not args.pretend:
         return ret
 
@@ -321,15 +354,25 @@ def run_shell(args):
 def extract(args):
     images = get_list(args)
     docker_argv = check_docker()
+    os.makedirs(args.output_dir, exist_ok=True)
     for imageid in images:
+        dest_dir = os.path.join(args.output_dir, imageid.replace(":","-"))
+        if os.path.exists(dest_dir):
+            print(f"Warning: destination '{dest_dir}' already exists.")
+            print(f"CTRL+C to abort; ENTER to DELETE '{dest_dir}'")
+            input()
+            shutil.rmtree(dest_dir)
         image = f"{DOCKER_NS}/{imageid}"
         info(image)
         prepare_image(args, docker_argv, image)
-        container = subprocess.check_output(docker_argv + ["create", image]).decode().strip()
-        subprocess.check_call(docker_argv + ["cp", f"{container}:/solvers/", args.output_dir])
-        os.rename(os.path.join(args.output_dir, "solvers"),
-                os.path.join(args.output_dir, imageid.replace(":","-")))
-        check_cmd(docker_argv + ["rm", container])
+        argv = docker_argv + ["run", "--rm", "-w", "/", "--entrypoint", "tar",
+                image,  "-c", "solvers"]
+        info(" ".join(argv))
+        with subprocess.Popen(argv, stdout=subprocess.PIPE,
+                stderr=sys.stderr) as p:
+            with tarfile.open(mode="r|", fileobj=p.stdout) as t:
+                t.extractall(args.output_dir)
+        os.rename(os.path.join(args.output_dir, "solvers"), dest_dir)
 
 #
 ##
@@ -382,16 +425,11 @@ def build_images(args):
         image = ImageManager(name, repo)
         setup = image.setup
 
-        fmtvars = {
-            "SOLVER": image.solver,
-            "SOLVER_NAME": image.solver_name,
-        }
-
         root = str(image.entry)
 
-        build_args = {k: v.format(**fmtvars) for k,v in setup.items() if \
+        build_args = {k: v.format(**image.vars) for k,v in setup.items() if \
                 k not in hide_opts and isinstance(v, str)}
-        build_args.update(fmtvars)
+        build_args.update(image.vars)
 
         builder_path = setup["builder"]
         if not builder_path.startswith("generic/"):
@@ -535,6 +573,12 @@ def main(redirected=False):
     docker_parser.add_argument("-e", "--env", action="append",
             help="(Docker option) Set environment variables")
     _docker_opts.append("env")
+
+    run_parser = argparse.ArgumentParser(add_help=False)
+    run_parser.add_argument("--timeout", type=int, default=3600,
+            help="Timeout for solver (in seconds; default: 3600)")
+    run_parser.add_argument("--fail-if-timeout", action="store_true",
+            help="Fail if timeout occurs")
     #
     ##
 
@@ -556,7 +600,7 @@ def main(redirected=False):
 
     p = subparsers.add_parser("run",
             help=f"Run one or several {DOCKER_NS} Docker images",
-            parents=[spec_parser, docker_parser])
+            parents=[spec_parser, run_parser, docker_parser])
     p.add_argument("dimacs",
             help="DIMACS file (possibly gzipped)")
     p.add_argument("proof", nargs="?",
@@ -565,7 +609,7 @@ def main(redirected=False):
 
     p = subparsers.add_parser("run-raw",
             help=f"Run one or several {DOCKER_NS} Docker images with direct call to solvers",
-            parents=[spec_parser, docker_parser])
+            parents=[spec_parser, run_parser, docker_parser])
     p.add_argument("args", nargs=argparse.REMAINDER,
             help="Arguments to docker image")
     p.set_defaults(func=runraw_images)
@@ -598,7 +642,7 @@ def main(redirected=False):
 
         p = subparsers.add_parser("test",
                 help=f"Test {DOCKER_NS} Docker images",
-                parents=[spec_parser, docker_parser])
+                parents=[spec_parser, run_parser, docker_parser])
         p.add_argument("--cnf", "-f", default="tests/cmu-bmc-barrel6.cnf.gz")
         p.set_defaults(func=test_images)
 
